{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE NamedFieldPuns        #-}
{-# LANGUAGE ScopedTypeVariables   #-}
{-# OPTIONS_GHC -Wall #-}

module Walker where

import           Control.Monad.Reader
import           SCAD
import           Shared


type LegName = String
type CenterOfMass = Position
type PosOffset = Position

data PackageConfig = PackageConfig {
        pkg_name :: String,
        scad_dir :: Filename,
        pkg_dir  :: Filename,
        src_dir  :: Filename
    };

type PackageMonad a = Reader PackageConfig a

data SCADModel = SCADModel {
        module_name :: SCADName,
        args        :: [Arg],
        imports     :: [Filename]
    } deriving (Show);

data ExtendedSCADModel = ExtendedSCADModel {
        smodel           :: SCADModel,
        scad_filename    :: Filename,
        stl_filename     :: Filename,
        urdf_filename    :: Filename,
        absolute_imports :: [Filename]
    } deriving (Show);

data Part m = Part {
            part_name        :: String,
            model            :: m,
            collision_models :: [m],
            part_weight      :: Maybe Mass,
            center_of_mass   :: Maybe CenterOfMass
        } deriving (Show);

type SimplePart = Part SCADModel
type ExtendedPart = Part ExtendedSCADModel

data Link p = Link {
            link_name :: String,
            parts     :: [p]
        } deriving (Show);

type SimpleLink = Link SimplePart
data ExtendedLink = ExtendedLink {
        link      :: Link ExtendedPart,
        urdf_name :: String
    } deriving (Show);

data JointType = RevoluteJoint {
            axis          :: Axis,
            min_angle     :: Angle,
            default_angle :: Angle,
            max_angle     :: Angle,
            vel_lim       :: AngularVelocity,
            effor_lim     :: Effort,
            g_steps       :: Integer
    } | FixedJoint {} deriving (Show);

data Joint l sub_j= Joint {
            joint_name :: String,
            offset     :: Pose,
            jtype      :: JointType,
            sub_link   :: l,
            sub_joint  :: Maybe sub_j
    } deriving (Show);

newtype SimpleJoint = SimpleJoint {
        j :: Joint SimpleLink SimpleJoint
    } deriving (Show);
data ExtendedJoint = ExtendedJoint {
            joint     :: Joint ExtendedLink ExtendedJoint,
            urdf_name :: String
    } deriving (Show);

data Leg j = Leg {
            leg_name   :: LegName,
            root_joint :: j
        } deriving (Show);

type SimpleLeg = Leg SimpleJoint
type ExtendedLeg = Leg ExtendedJoint

data Body p = Body {
        body_part :: p,
        body_name :: String
    } deriving (Show);

type SimpleBody = Body SimplePart
data ExtendedBody = ExtendedBody {
    base_body :: Body ExtendedPart,
    urdf_name :: String
} deriving (Show);

data Walker l b = Walker {
            walker_name :: String,
            legs        :: [l],
            body        :: b
    } deriving (Show);

type SimpleWalker = Walker SimpleLeg SimpleBody
type ExtendedWalker = Walker ExtendedLeg ExtendedBody


extendWalker :: SimpleWalker -> PackageMonad ExtendedWalker
extendWalker w = do
    elegs <- traverse extend_leg $ legs w
    ebody <- extend_body $ body w
    return $  w {
            legs = elegs,
            body = ebody
        }
    where
        extend_model :: String -> SCADModel -> PackageMonad ExtendedSCADModel
        extend_model prefix m = do
            scad <- asks scad_dir
            pkg_d <- asks pkg_dir
            src <- asks src_dir
            pkg <- asks pkg_name
            let stl = "stl/" ++ prefix ++ "/" ++ module_name m ++ ".stl"
            return $ ExtendedSCADModel {
                smodel = m,
                scad_filename = scad ++ prefix ++ "/" ++ module_name m ++ ".scad",
                stl_filename = pkg_d ++ "/" ++ stl,
                urdf_filename = "package://" ++ pkg ++ "/" ++ stl,
                absolute_imports = (\x -> src ++ "/" ++ x) <$> imports m
            }

        extend_part :: String -> SimplePart -> PackageMonad ExtendedPart
        extend_part prefix p = do
            model <- extend_model prefix $ model p
            col_models <- traverse (extend_model prefix) $ collision_models p
            return $ p {
                model = model,
                collision_models = col_models
            }

        extend_link :: SimpleLeg -> SimpleLink -> PackageMonad ExtendedLink
        extend_link leg l = do
            parts <- traverse (extend_part urdf_name) $ parts l
            return $ ExtendedLink {
                link = l { parts = parts },
                urdf_name = urdf_name
            }
            where
                urdf_name = "leg/" ++ leg_name leg ++ "/" ++ link_name l

        extend_joint :: SimpleLeg -> SimpleJoint -> PackageMonad ExtendedJoint
        extend_joint leg SimpleJoint{j} = do
            sub_l <- extend_link leg $ sub_link j
            sub_jnt <- case sub_joint j of
                (Just child) -> do
                    echild <- extend_joint leg child
                    return $ Just echild
                Nothing -> return Nothing
            return $ ExtendedJoint {
                joint = j {
                    sub_link = sub_l,
                    sub_joint = sub_jnt
                },
                urdf_name = "leg/" ++ leg_name leg  ++ "/" ++ joint_name j
            }

        extend_leg :: SimpleLeg -> PackageMonad ExtendedLeg
        extend_leg l = do
            rjoint <- extend_joint l $ root_joint l
            return $ l { root_joint = rjoint }

        extend_body :: SimpleBody -> PackageMonad ExtendedBody
        extend_body b = do
            bpart <- extend_part (body_name b) $ body_part b
            return $ ExtendedBody {
                base_body = b { body_part = bpart },
                urdf_name = body_name b
            }


foldJoints :: (ExtendedJoint -> a -> a) -> a -> ExtendedJoint -> a
foldJoints f base ej = case sub_joint $ joint ej of
            Just child -> f ej $ foldJoints f base child
            Nothing    -> f ej base

mapJoints :: (ExtendedJoint -> a) -> ExtendedJoint -> [a]
mapJoints f = foldJoints (\w x -> f w:x) []

mapLinks :: (ExtendedLink -> a) -> ExtendedJoint -> [a]
mapLinks f = mapJoints ( f . sub_link . joint )

mapParts :: forall a . (ExtendedPart -> a) -> ExtendedWalker -> [a]
mapParts f Walker{legs, body} = f (body_part $ base_body body) : concatMap map_legs legs
    where
        map_legs :: ExtendedLeg -> [a]
        map_legs Leg {root_joint} = concat $ mapLinks f_part root_joint

        f_part :: ExtendedLink -> [a]
        f_part x = f <$> parts (link x)

mapModels :: forall a . (ExtendedSCADModel -> a) -> ExtendedWalker -> [a]
mapModels f walker = concat $ mapParts part_to_models walker
    where
        part_to_models :: ExtendedPart -> [a]
        part_to_models Part{model, collision_models} = f model : (f <$> collision_models)



