{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE NamedFieldPuns        #-}

module Walker.URDF where

import           Data.Hashable                     (hash)
import           Linear.V3
import qualified Numeric.Units.Dimensional.Prelude as U
import           Shared
import qualified URDF
import           Walker


poseToURDFOrigin :: Pose -> URDF.Origin
poseToURDFOrigin pose = URDF.Origin (Just $ pos pose)
                                    (Just $ orientation rpy)
 where
  rpy = (U.*~ U.radian) <$> quatToRPY (ori pose)
  orientation (V3 r p y) = URDF.Orientation r p y

getURDFLegJoints :: ExtendedBody -> ExtendedLeg -> [URDF.Joint]
getURDFLegJoints body leg = link_to_joint <$> triplets
 where
    joints :: [ExtendedJoint]
    joints = mapJoints id $ root_joint leg

    link_names :: [String]
    link_names =  mapLinks (\x -> urdf_name ( x :: ExtendedLink)) $ root_joint leg

    triplets = zip3 link_names joints (tail link_names)

    link_to_joint :: (String, ExtendedJoint, String) -> URDF.Joint
    link_to_joint (p_link_name, ExtendedJoint { urdf_name, joint }, ch_name) = URDF.Joint {
        name = urdf_name,
        jtype = case jtype joint of
            RevoluteJoint {axis, min_angle, max_angle, vel_lim, effor_lim}
                -> URDF.Revolute {
                    limit = URDF.Limit effor_lim vel_lim (Just min_angle) (Just max_angle),
                    axis = Just axis,
                    rev_dynamics = Nothing
                }
            FixedJoint
                -> URDF.Fixed {},
        parent = p_link_name,
        child = ch_name,
        origin = Just $ poseToURDFOrigin $ offset joint,
        calibration = Nothing,
        mimic = Nothing,
        safety_controller = Nothing
    }

getURDFPartVisual :: String -> ExtendedPart -> URDF.Visual
getURDFPartVisual link_name Part{model} = URDF.Visual
      { v_geometry = URDF.Mesh (urdf_filename model)
                               (Just $ V3 0.001 0.001 0.001)
      , name       = Nothing
      , origin     = Nothing
      , material   = Just $ URDF.Material link_name (Just color) Nothing
      }
 where
   color = URDF.Color
    { r = hash_color (link_name ++ "red")
    , g = hash_color (link_name ++ "green")
    , b = hash_color (link_name ++ "blue")
    , a = 1
    }

   hash_color :: String -> ColorChannel
   hash_color seed = fromIntegral (hash seed `mod` 1024) / 1024.0


getURDFPartCollisions :: ExtendedPart -> [URDF.Collision]
getURDFPartCollisions Part {collision_models} = f <$> collision_models
 where
  f ExtendedSCADModel {urdf_filename} = URDF.Collision
        { geometry = URDF.Mesh urdf_filename
                               (Just $ V3 0.001 0.001 0.001)
        , name     = Nothing
        , origin   = Nothing
        }


getURDFLegLinks :: ExtendedLeg -> [URDF.Link]
getURDFLegLinks leg = mapLinks link_to_urdf $ root_joint leg
 where
  link_to_urdf ExtendedLink{urdf_name, link} = URDF.Link { name      = urdf_name
        , visuals   = getURDFPartVisual urdf_name <$> parts link
        , inertial  = Nothing
        , collision = concat $ getURDFPartCollisions <$> parts link
        }

getURDFBodyLink :: ExtendedBody -> URDF.Link
getURDFBodyLink ExtendedBody {base_body, urdf_name} =  URDF.Link
      { name      = urdf_name
      , visuals   = [ getURDFPartVisual urdf_name $ body_part base_body ]
      , inertial  = Nothing
      , collision = getURDFPartCollisions $ body_part base_body
      }

getURDF :: ExtendedWalker -> URDF.URDF
getURDF walker = URDF.URDF {
                name = walker_name walker,
                joints = concatMap (getURDFLegJoints $ body walker ) $ legs walker,
                links = getURDFBodyLink (body walker) : (concatMap getURDFLegLinks $ legs walker)
            }
